/* From https://github.com/bos/criterion */

#include <time.h>

void unliftio_inittime(void)
{
}

double unliftio_gettime(void)
{
    struct timespec ts;

    clock_gettime(CLOCK_MONOTONIC, &ts);

    return ts.tv_sec + ts.tv_nsec * 1e-9;
}
